import logging
import os
from typing import Optional


def get_logger(name: str,
               log_file: Optional[str] = None,
               level: int = logging.INFO,
               console_level: int = logging.INFO,
               file_level: int = logging.INFO,
               mode: str = 'w') -> logging.Logger:
    """
    Function to create a logger with a given name and log to both console and file.

    Args:
        name (str): The name of the logger.
        log_file (str, optional): The file path to write the log output. Defaults to None.
        level (int): The threshold level for the logger. Defaults to logging.INFO.
        console_level (int): The threshold level for logging output to the console. Defaults to logging.INFO.
        file_level (int): The threshold level for logging output to the file. Defaults to logging.INFO.
        mode (str): The mode in which the file is opened. Defaults to 'w' - write, can be 'a' - append.

    Raises:
        FileNotFoundError: If the directory specified in 'log_file' does not exist.

    Returns:
        logging.Logger: A configured logger object that logs to both the file and console.

    Example:
        logger = get_logger('example_logger', 'logs/app.log', level=logging.DEBUG, console_level=logging.INFO, file_level=logging.ERROR)
    """
    # Create logger
    logger = logging.getLogger(name)
    logger.setLevel(level)

    # Avoid adding multiple handlers if function is called multiple times
    if logger.hasHandlers():
        logger.handlers.clear()

    # Create formatter
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    # Check if log_file is provided and is a valid path
    if log_file:
        # Check if the directory for the log_file exists, if not raise an error
        log_dir = os.path.dirname(log_file)
        if not os.path.exists(log_dir):
            raise FileNotFoundError(f"The directory '{log_dir}' does not exist. Cannot create log file.")
        # Create a file handler and set level and formatter
        file_handler = logging.FileHandler(log_file, mode=mode)
        file_handler.setLevel(file_level)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    # Always create and add a console handler
    console_handler = logging.StreamHandler()
    console_handler.setLevel(console_level)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    logger.propagate = False

    return logger



import datetime
import gc
import os
import threading
import time

import cv2

from openmmla_vision.utils.logger_utils import get_logger


class WebcamVideoStream:
    logger = get_logger('WebcamVideoStream')

    def __init__(self, format, res, src=0, save_path=''):
        self.stream = cv2.VideoCapture(src)
        self.stream.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc(*format))
        self.stream.set(3, res[0])  # Width
        self.stream.set(4, res[1])  # Height
        self.stream.set(cv2.CAP_PROP_AUTOFOCUS, 0)
        self.grabbed, self.frame = self.stream.read()
        self.save_path = save_path

        # Determine the type of save files, either video or images
        if self.save_path:
            if any(self.save_path.lower().endswith(ext) for ext in ['.avi', '.mp4']):
                fourcc = cv2.VideoWriter_fourcc(*format)
                self.video_writer = cv2.VideoWriter(self.save_path, fourcc, 30.0, res)
            else:
                if not os.path.exists(self.save_path):
                    os.makedirs(self.save_path)

        self.stop_event = threading.Event()
        self.threads = []

    def start(self):
        self.stop_event.clear()
        self._start_thread(self._update)
        return self

    def stop(self):
        self.stop_event.set()
        for t in self.threads:
            if threading.current_thread() != t:
                try:
                    t.join(timeout=5)
                except Exception as e:
                    self.logger.warning("During the thread stopping, catch exception %s", e, exc_info=True)
        self.threads.clear()
        self._free_memory()

    def read(self):
        return self.frame

    def _start_thread(self, target, *args):
        t = threading.Thread(target=target, args=args)
        t.daemon = True
        self.threads.append(t)
        t.start()

    def _update(self):
        start_time = time.time()
        fps = 0
        frame_id = 0
        frame_count = 0

        while not self.stop_event.is_set():
            (self.grabbed, self.frame) = self.stream.read()
            if not self.grabbed:
                raise Exception('Frame not grabbed')

            # Display timestamp on frame
            now = datetime.datetime.now()
            current_time_str = now.strftime("%Y-%m-%d %H:%M:%S")
            cv2.putText(self.frame, current_time_str, (self.frame.shape[1] - 300, 30), cv2.FONT_HERSHEY_SIMPLEX,
                        0.7, (0, 255, 0), 2)

            # Display fps on frame
            frame_count += 1
            if frame_count % 30 == 0:
                elapsed_time = time.time() - start_time
                fps = frame_count / elapsed_time
                start_time = time.time()
                frame_count = 0
                # print(f'Webcam FPS: {fps:.2f}')  # Uncomment to print fps for testing
            cv2.putText(self.frame, f"FPS: {fps:.2f}", (10, 30), cv2.FONT_HERSHEY_SIMPLEX, 0.7, (0, 255, 0),
                        2)

            if self.save_path:
                if hasattr(self, 'video_writer'):
                    self.video_writer.write(self.frame)
                else:
                    saved_frame = cv2.resize(self.frame, (960, 540))
                    cv2.imwrite(os.path.join(self.save_path, f'frame_{frame_id:07d}.jpg'), saved_frame)
                    frame_id += 1

    def _free_memory(self):
        if hasattr(self, 'video_writer'):
            self.video_writer.release()
        self.stream.release()
        gc.collect()

    def __enter__(self):
        return self.start()

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()



import logging
import threading


class RaisingThread(threading.Thread):
    def run(self):
        self._exc = None
        try:
            super().run()
        except Exception as e:
            self._exc = e
            logging.error(f"Error in thread {self.name}: {e}", exc_info=True)

    def join(self, timeout=None):
        super().join(timeout=timeout)
        if self._exc:
            print(f"Raising exception {self._exc} from thread {self.name}")
            raise self._exc



import configparser

import influxdb_client
from influxdb_client.client.write_api import SYNCHRONOUS


class InfluxDBClientWrapper(influxdb_client.InfluxDBClient):
    """Extended InfluxDB client that loads configuration from a file and adds custom functionalities."""

    def __init__(self, config_path):
        """Initialize InfluxDB client with configurations"""
        config = configparser.ConfigParser()
        config.read(config_path)

        super().__init__(url=config['InfluxDB']['url'], token=config['InfluxDB']['token'],
                         org=config['InfluxDB']['org'])

        # Initialize APIs
        self.query_api = self.query_api()
        self.write_api = self.write_api(write_options=SYNCHRONOUS)
        self.bucket_api = self.buckets_api()

    def query(self, query):
        """Query data from InfluxDB"""
        return self.query_api.query(query)

    def write(self, bucket, record):
        """Write data to InfluxDB"""
        return self.write_api.write(bucket=bucket, record=record)

    def get_buckets(self):
        """Get all buckets from InfluxDB"""
        return self.bucket_api.find_buckets()

    def create_bucket(self, bucket_name):
        """Create a new bucket in InfluxDB"""
        return self.bucket_api.create_bucket(bucket_name=bucket_name)



import select
import sys
from datetime import datetime


def flush_input():
    """Flush all input from stdin buffer."""
    while sys.stdin in select.select([sys.stdin], [], [], 0)[0]:
        sys.stdin.read(1)


def get_bucket_name(influx_client) -> str:
    """Get the bucket name from user input or create a new bucket if needed"""
    bucket_name = None

    while True:
        flush_input()
        print("------------------------------------------------")
        bucket_list = influx_client.get_buckets()
        bucket_names = [bucket.name for bucket in bucket_list.buckets if bucket.name not in ['_tasks', '_monitoring']]

        try:
            bucket_names = sorted(bucket_names, key=lambda x: datetime.strptime(x.split('_')[1], '%Y-%m-%dT%H:%M:%SZ'))
            for i, name in enumerate(bucket_names, start=1):
                print(f"{i}. {name}")
            bucket_idx = input("Enter the number of the bucket you want to enroll in, or enter 'n' for a new bucket, "
                               "or 'r' to refresh the list: ")
        except Exception as e:
            print(f'No compatible bucket session to sort, {e}')
            bucket_idx = 'n'

        if bucket_idx.isdigit() and 1 <= int(bucket_idx) <= len(bucket_names):
            bucket_name = bucket_names[int(bucket_idx) - 1]
            print(f"Bucket: {bucket_name} has been selected.")
            break
        elif bucket_idx in ['n', 'new']:
            timestamp = datetime.utcnow().isoformat().split('.')[0] + 'Z'
            bucket_name = 'session_' + timestamp
            influx_client.create_bucket(bucket_name=bucket_name)
            print(f"Bucket: {bucket_name} has been created.")
            break
        elif bucket_idx in ['r', 'refresh', '']:
            continue
        else:
            print("Invalid input. Please enter a number from the list, or 'n' for a new bucket, or 'r' to refresh the "
                  "list")

    return bucket_name


def get_function_calibrator():
    """Get the function to performed from user input for camera calibrator"""
    while True:
        try:
            flush_input()
            print("------------------------------------------------")
            select_fun = input(f"Please input your operation:\n"
                                 f"1: capture chess pattern images from you camera\n"
                                 f"2: calibrate your selected camera\n"
                                 f"0: exit\n"
                                 f"Selected function: ")
            if select_fun.strip():
                return int(select_fun)
            else:
                print('Please enter a value')
        except EOFError:
            print(
                "\nUnexpected input received. If you resized the terminal or pressed certain keys, please avoid doing "
                "so and try again.")
        except ValueError:
            print('Please enter a valid integer')

def get_function_sync_manager():
    """Get the function to performed from user input for camera coordinates synchronizer"""
    while True:
        try:
            flush_input()
            print("------------------------------------------------")
            select_fun = input(f"Please input your operation:\n"
                               f"1: start\n"
                               f"2: switch mode\n"
                               f"3: set camera id\n"
                               f"4: export transformations\n"
                               f"5: clear transformations\n"
                               f"0: exit\n"
                               f"Selected function: ")
            if select_fun.strip():
                return int(select_fun)
            else:
                print('Please enter a value')
        except EOFError:
            print(
                "\nUnexpected input received. If you resized the terminal or pressed certain keys, please avoid doing "
                "so and try again.")
        except ValueError:
            print('Please enter a valid integer')


def get_function_base():
    """Get the function to be performed from user input for video base"""
    while True:
        try:
            flush_input()
            print("------------------------------------------------")
            select_fun = input(f"Please input your operation:\n"
                               f"1: start\n"
                               f"2: set camera\n"
                               f"0: exit\n"
                               f"Selected function: ")

            if select_fun.strip():  # Check if input is not empty after removing leading/trailing whitespace
                return int(select_fun)
            else:
                print('Please enter a value')

        except EOFError:
            print(
                "\nUnexpected input received. If you resized the terminal or pressed certain keys, please avoid doing "
                "so and try again.")

        except ValueError:
            print('Please enter a valid integer')


def get_function_synchronizer():
    """Get the function to be performed from user input for synchronizer"""
    while True:
        try:
            flush_input()
            print("------------------------------------------------")
            select_fun = input(f"Please input your operation:\n"
                               f"1: start\n"
                               f"2: set main camera\n"
                               f"0: exit\n"
                               f"Selected function: ")

            if select_fun.strip():  # Check if input is not empty after removing leading/trailing whitespace
                return int(select_fun)
            else:
                print('Please enter a value')

        except EOFError:
            print(
                "\nUnexpected input received. If you resized the terminal or pressed certain keys, please avoid doing "
                "so and try again.")

        except ValueError:
            print('Please enter a valid integer')


def get_function_visualizer():
    """Get the function to be performed from user input for visualizer"""
    while True:
        try:
            flush_input()
            print("------------------------------------------------")
            select_fun = input(f"Please input your operation:\n"
                               f"1: start\n"
                               f"0: exit\n"
                               f"Selected function: ")

            if select_fun.strip():  # Check if input is not empty after removing leading/trailing whitespace
                return int(select_fun)
            else:
                print('Please enter a value')

        except EOFError:
            print(
                "\nUnexpected input received. If you resized the terminal or pressed certain keys, please avoid doing "
                "so and try again.")

        except ValueError:
            print('Please enter a valid integer')



import json
import os
from collections import defaultdict

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
from pyecharts import options as opts
from pyecharts.charts import Line
from pyecharts.charts import Page
from pyecharts.commons.utils import JsCode
from scipy.interpolate import interp1d
from scipy.spatial.distance import euclidean

BADGE_COLOR_MAP = {
    '0': 'red',
    '1': 'blue',
    '2': 'green',
    '3': 'purple',
    '4': 'orange',
    '5': 'pink',
    '6': 'brown',
    '7': 'cyan',
    '8': 'magenta',
    '9': 'yellow',
    '10': 'black',
    '11': 'lime',
    '12': 'maroon',
    '13': 'olive',
    '14': 'navy',
    '15': 'teal',
    '16': 'gray',
    '17': 'sienna',
    '18': 'violet',
    '19': 'turquoise'
}
CANDIDATES = [str(i) for i in range(20)]


def session_analysis_video(project_dir, bucket_name, influx_client):
    """Retrieves data from InfluxDB for badge relations, translations, and rotations, and then visualizes them."""
    logs_dir = os.path.join(project_dir, 'logs')
    log_dir = os.path.join(logs_dir, f'{bucket_name}')
    visualizations_dir = os.path.join(project_dir, 'visualizations')
    visualization_dir = os.path.join(visualizations_dir, f"{bucket_name}/post-time")
    os.makedirs(log_dir, exist_ok=True)
    os.makedirs(visualization_dir, exist_ok=True)

    # Log
    relations_data = fetch_and_process_data(bucket_name, "badge relations", influx_client)
    translations_data = fetch_and_process_data(bucket_name, "badge translations", influx_client)
    rotations_data = fetch_and_process_data(bucket_name, "badge rotations", influx_client)
    translation_json_file = save_to_json_file(bucket_name, translations_data, "badge_translations", log_dir)
    relations_json_file = save_to_json_file(bucket_name, relations_data, "badge_relations", log_dir)
    save_to_json_file(bucket_name, rotations_data, "badge_rotations", log_dir)

    # Visualize
    plot_badge_locations_and_trajectories(translation_json_file, visualization_dir)
    plot_2d_heatmap(translation_json_file, visualization_dir)
    plot_physical_interaction_network(relations_json_file, visualization_dir)

    # Analyze across sessions
    across_sessions_analysis_video(logs_dir, visualizations_dir)


def fetch_and_process_data(bucket_name, measurement, influx_client):
    """Queries InfluxDB for specified data, converts it to JSON, and sorts it based on 'segment_start_time'."""
    query = generate_query(bucket_name, measurement)
    tables = influx_client.query(query)
    json_str = tables.to_json(indent=5)
    data = json.loads(json_str)
    data.sort(key=lambda x: x['segment_start_time'])
    return json.dumps(data, indent=5)


def generate_query(bucket_name, measurement):
    """Constructs an InfluxDB query string for a specific bucket, measurement, and fields."""
    bucket_start_time = bucket_name.split('_')[1]
    return f"""from(bucket: "{bucket_name}")
                |> range(start: {bucket_start_time})
                |> filter(fn: (r) => r._measurement == "{measurement}")
                |> pivot(rowKey:["_time"], columnKey: ["_field"], valueColumn: "_value")"""


def read_json_file(file_path):
    """Reads and returns the contents of a JSON file."""
    try:
        with open(file_path, 'r') as f:
            return json.load(f)
    except Exception as e:
        print(f"An error occurred while reading the JSON file: {e}")
        return None


def save_to_json_file(bucket_name, data, suffix, log_dir):
    """Saves the given data to a JSON file in a specified directory, naming it based on the bucket name and a suffix."""
    json_path = os.path.join(log_dir, f"{bucket_name}_{suffix}.json")
    with open(json_path, 'w') as f:
        f.write(data)
    print(f"{suffix} saved to {bucket_name}_{suffix}.json")
    return json_path


def convert_to_dataframe(json_data, json_columns):
    """Converts JSON data into a pandas DataFrame and transforms specific JSON-formatted string columns into Python
    dictionaries."""
    df = pd.DataFrame(json_data)[json_columns]
    for column in json_columns:
        if column != 'segment_start_time':
            df[column] = df[column].apply(json.loads)
    return df


def weight_to_width(wt, min_wt=0.0, max_wt=0.5, min_width=1, max_width=20):
    """Converts a numerical weight value to a corresponding width value for visual representation in a graph."""
    wt = max(min_wt, min(max_wt, wt))
    width = min_width + (wt - min_wt) / (max_wt - min_wt) * (max_width - min_width)

    return width


# Credit for this method goes to Stackoverflow user kcoskun
# https://stackoverflow.com/questions/22785849/drawing-multiple-edges-between-two-nodes-with-networkx
def my_draw_networkx_edge_labels(G, pos, edge_labels=None, label_pos=0.5, font_size=10, font_color="k",
                                 font_family="sans-serif", font_weight="normal", alpha=None, bbox=None,
                                 horizontalalignment="center", verticalalignment="center", ax=None, rotate=True,
                                 clip_on=True, rad=0):
    if ax is None:
        ax = plt.gca()
    if edge_labels is None:
        labels = {(u, v): d for u, v, d in G.edges(data=True)}
    else:
        labels = edge_labels
    text_items = {}
    for (n1, n2), label in labels.items():
        (x1, y1) = pos[n1]
        (x2, y2) = pos[n2]
        (x, y) = (
            x1 * label_pos + x2 * (1.0 - label_pos),
            y1 * label_pos + y2 * (1.0 - label_pos),
        )
        pos_1 = ax.transData.transform(np.array(pos[n1]))
        pos_2 = ax.transData.transform(np.array(pos[n2]))
        linear_mid = 0.5 * pos_1 + 0.5 * pos_2
        d_pos = pos_2 - pos_1
        rotation_matrix = np.array([(0, 1), (-1, 0)])
        ctrl_1 = linear_mid + rad * rotation_matrix @ d_pos
        ctrl_mid_1 = 0.5 * pos_1 + 0.5 * ctrl_1
        ctrl_mid_2 = 0.5 * pos_2 + 0.5 * ctrl_1
        bezier_mid = 0.5 * ctrl_mid_1 + 0.5 * ctrl_mid_2
        (x, y) = ax.transData.inverted().transform(bezier_mid)

        if rotate:
            # in degrees
            angle = np.arctan2(y2 - y1, x2 - x1) / (2.0 * np.pi) * 360
            # make label orientation "right-side-up"
            if angle > 90:
                angle -= 180
            if angle < -90:
                angle += 180
            # transform data coordinate angle to screen coordinate angle
            xy = np.array((x, y))
            trans_angle = ax.transData.transform_angles(
                np.array((angle,)), xy.reshape((1, 2))
            )[0]
        else:
            trans_angle = 0.0
        # use default box of white with white border
        # if bbox is None:
        #     bbox = dict(boxstyle="round", ec=(1.0, 1.0, 1.0), fc=(1.0, 1.0, 1.0))
        if not isinstance(label, str):
            label = str(label)  # this makes "1" and 1 labeled the same

        t = ax.text(
            x,
            y,
            label,
            size=font_size,
            color=font_color,
            family=font_family,
            weight=font_weight,
            alpha=alpha,
            horizontalalignment=horizontalalignment,
            verticalalignment=verticalalignment,
            rotation=trans_angle,
            transform=ax.transData,
            bbox=bbox,
            zorder=1,
            clip_on=clip_on,
        )
        text_items[(n1, n2)] = t

    ax.tick_params(
        axis="both",
        which="both",
        bottom=False,
        left=False,
        labelbottom=False,
        labelleft=False,
    )

    return text_items


def plot_badge_locations_and_trajectories(json_file_path, visualization_dir, plot_type='both'):
    """Visualizes the 2D and 3D locations and trajectories of badges from JSON data."""
    bucket_name = f"session_{os.path.basename(json_file_path).split('_')[1]}"

    # Load the JSON file
    with open(json_file_path, 'r') as f:
        json_data = json.load(f)

    # Initialize a dictionary to store the badge locations
    badge_translations = defaultdict(list)
    badge_time_stamps = defaultdict(list)

    # Extract and store badge locations from the JSON data
    for entry in json_data:
        coords_dict = json.loads(entry['translations'])
        segment_time = entry['segment_start_time']
        for badge, coords in coords_dict.items():
            if badge in CANDIDATES:
                x, y, z = coords
                badge_translations[badge].append({
                    'x': x[0],
                    'y': y[0],
                    'z': z[0]
                })
                badge_time_stamps[badge].append(segment_time)

    if plot_type in ['both', 'position']:
        # 2D scatter plot of positions
        plt.figure(figsize=(12, 8))

        # Plot the positions in the order of the color map
        for badge in BADGE_COLOR_MAP.keys():
            if badge in badge_translations:
                z_coords = [entry['z'] for entry in badge_translations[badge]]
                x_coords = [entry['x'] for entry in badge_translations[badge]]
                plt.scatter(z_coords, x_coords, s=1, color=BADGE_COLOR_MAP[badge], label=f"Badge {badge}")

        plt.xlabel('Z Coordinate')
        plt.ylabel('X Coordinate')
        plt.title('2D Localisation of Badges (Z, X)')
        plt.legend()
        plt.savefig(os.path.join(visualization_dir, f"{bucket_name}_2d_positions.png"), dpi=300)
        print(f"Saved 2D positions to {bucket_name}_2d_positions.png")

    if plot_type in ['both', 'trajectory']:
        # 2D scatter plot with interpolated trajectory lines
        plt.figure(figsize=(12, 8))

        # Plot dummy scatter points for legend
        for badge in BADGE_COLOR_MAP.keys():
            if badge in badge_translations:
                plt.scatter([], [], color=BADGE_COLOR_MAP[badge], label=f"Badge {badge} Trajectory")

        # Plot the actual trajectories
        for badge, coords_list in badge_translations.items():
            z_coords = np.array([entry['z'] for entry in coords_list])
            x_coords = np.array([entry['x'] for entry in coords_list])
            time_stamps = badge_time_stamps[badge]
            # Initialize start point for the new segment
            start_idx = 0

            for idx in range(1, len(time_stamps)):
                # Check if the time difference exceeds 1.5 seconds
                if time_stamps[idx] - time_stamps[idx - 1] > 1.5:
                    # Interpolate and plot the current segment
                    segment_z = z_coords[start_idx:idx]
                    segment_x = x_coords[start_idx:idx]
                    if len(set(time_stamps[start_idx:idx])) < 2:
                        plt.plot(z_coords[start_idx], x_coords[start_idx], linestyle=':', marker=None,
                                 color=BADGE_COLOR_MAP.get(badge, 'black'))
                        continue
                    # Determine interpolation type
                    interpolation_type = 'cubic' if len(segment_z) >= 4 else 'linear'
                    t = np.linspace(0, 1, len(segment_z))
                    interpolator = interp1d(t, (segment_z, segment_x), kind=interpolation_type)
                    new_t = np.linspace(0, 1, max(500, len(segment_z)))
                    new_z_coords, new_x_coords = interpolator(new_t)
                    plt.plot(new_z_coords, new_x_coords, linestyle=':', marker=None,
                             color=BADGE_COLOR_MAP.get(badge, 'black'))
                    # Update start point for the new segment
                    start_idx = idx

            # Interpolate and plot the last segment
            if len(set(time_stamps[start_idx:])) >= 2:
                segment_z = z_coords[start_idx:]
                segment_x = x_coords[start_idx:]
                interpolation_type = 'cubic' if len(segment_z) >= 4 else 'linear'
                t = np.linspace(0, 1, len(segment_z))
                interpolator = interp1d(t, (segment_z, segment_x), kind=interpolation_type)
                new_t = np.linspace(0, 1, max(500, len(segment_z)))
                new_z_coords, new_x_coords = interpolator(new_t)
                plt.plot(new_z_coords, new_x_coords, linestyle=':', marker=None,
                         color=BADGE_COLOR_MAP.get(badge, 'black'))

            # Mark start and end points
            plt.scatter(z_coords[0], x_coords[0], color=BADGE_COLOR_MAP.get(badge, 'black'), marker='.',
                        zorder=5)
            plt.scatter(z_coords[-1], x_coords[-1], color=BADGE_COLOR_MAP.get(badge, 'black'), marker='X', s=20,
                        zorder=5)

        plt.xlabel('Z Coordinate')
        plt.ylabel('X Coordinate')
        plt.title('2D Interpolated Trajectories of Badges (Z, X)')
        plt.legend()
        plt.savefig(os.path.join(visualization_dir, f"{bucket_name}_2d_interpolated_trajectories.png"), dpi=300)
        print(f"Saved 2D interpolated trajectories to {bucket_name}_2d_interpolated_trajectories.png")

    if plot_type in ['both', 'position']:
        # 3D scatter plot of positions
        fig = plt.figure(figsize=(12, 10))
        ax = fig.add_subplot(111, projection='3d')

        # Plot the positions in the order of the color map
        for badge in BADGE_COLOR_MAP.keys():
            if badge in badge_translations:
                z_coords = [entry['z'] for entry in badge_translations[badge]]
                x_coords = [entry['x'] for entry in badge_translations[badge]]
                y_coords = [entry['y'] for entry in badge_translations[badge]]
                ax.scatter(z_coords, x_coords, y_coords, s=1, color=BADGE_COLOR_MAP[badge], label=f"Badge {badge}")

        ax.set_xlabel('Z Coordinate')
        ax.set_ylabel('X Coordinate')
        ax.set_zlabel('Y Coordinate')
        ax.set_title('3D Localisation of Badges')
        ax.legend()
        plt.savefig(os.path.join(visualization_dir, f"{bucket_name}_3d_positions.png"), dpi=300)
        print(f"Saved 3D positions to {bucket_name}_3d_positions.png")

    if plot_type in ['both', 'trajectory']:
        # 3D scatter plot with interpolated trajectory lines
        fig = plt.figure(figsize=(12, 10))
        ax = fig.add_subplot(111, projection='3d')

        # Plot dummy scatter points for legend
        for badge in BADGE_COLOR_MAP.keys():
            if badge in badge_translations:
                ax.scatter([], [], [], color=BADGE_COLOR_MAP[badge], label=f"Badge {badge} Trajectory")

        # Plot the actual trajectories
        for badge, coords_list in badge_translations.items():
            z_coords = np.array([entry['z'] for entry in coords_list])
            x_coords = np.array([entry['x'] for entry in coords_list])
            y_coords = np.array([entry['y'] for entry in coords_list])
            time_stamps = badge_time_stamps[badge]
            # Initialize start point for the new segment
            start_idx = 0

            for idx in range(1, len(time_stamps)):
                # Check if the time difference exceeds 1.5 seconds
                if time_stamps[idx] - time_stamps[idx - 1] > 1.5:
                    # Interpolate and plot the current segment
                    if len(set(time_stamps[start_idx:idx])) < 2:
                        ax.plot(z_coords[start_idx], x_coords[start_idx], y_coords[start_idx], linestyle=':',
                                color=BADGE_COLOR_MAP.get(badge, 'black'))
                        continue
                    # Determine interpolation type
                    interpolation_type = 'cubic' if idx - start_idx >= 4 else 'linear'
                    segment_z = z_coords[start_idx:idx]
                    segment_x = x_coords[start_idx:idx]
                    segment_y = y_coords[start_idx:idx]
                    t = np.linspace(0, 1, len(segment_z))
                    interpolator = interp1d(t, (segment_z, segment_x, segment_y), kind=interpolation_type)
                    new_t = np.linspace(0, 1, max(500, len(segment_z)))
                    new_z_coords, new_x_coords, new_y_coords = interpolator(new_t)
                    ax.plot(new_z_coords, new_x_coords, new_y_coords, linestyle=':',
                            color=BADGE_COLOR_MAP.get(badge, 'black'))
                    # Update start point for the new segment
                    start_idx = idx

            # Interpolate and plot the last segment
            if len(set(time_stamps[start_idx:])) >= 2:
                segment_z = z_coords[start_idx:]
                segment_x = x_coords[start_idx:]
                segment_y = y_coords[start_idx:]
                interpolation_type = 'cubic' if len(segment_z) >= 4 else 'linear'
                t = np.linspace(0, 1, len(segment_z))
                interpolator = interp1d(t, (segment_z, segment_x, segment_y), kind=interpolation_type)
                new_t = np.linspace(0, 1, max(500, len(segment_z)))
                new_z_coords, new_x_coords, new_y_coords = interpolator(new_t)
                ax.plot(new_z_coords, new_x_coords, new_y_coords, linestyle=':',
                        color=BADGE_COLOR_MAP.get(badge, 'black'))

            # Mark start and end points
            ax.scatter(z_coords[0], x_coords[0], y_coords[0], color=BADGE_COLOR_MAP.get(badge, 'black'),
                       marker='.', s=50, zorder=5)
            ax.scatter(z_coords[-1], x_coords[-1], y_coords[-1], color=BADGE_COLOR_MAP.get(badge, 'black'),
                       marker='X', s=20, zorder=5)

        ax.set_xlabel('Z Coordinate')
        ax.set_ylabel('X Coordinate')
        ax.set_zlabel('Y Coordinate')
        ax.set_title('3D Interpolated Trajectories of Badges')
        ax.legend()
        plt.savefig(os.path.join(visualization_dir, f"{bucket_name}_3d_interpolated_trajectories.png"), dpi=300)
        print(f"Saved 3D interpolated trajectories to {bucket_name}_3d_interpolated_trajectories.png")


def plot_2d_heatmap(json_file_path, visualization_dir):
    """Creates a 2D heatmap visualization of badge positions based on their Z and X coordinates."""
    bucket_name = f"session_{os.path.basename(json_file_path).split('_')[1]}"

    # Load the JSON file to inspect its structure
    with open(json_file_path, 'r') as f:
        json_data = json.load(f)

    # Initialize a dictionary to store the badge translations
    badge_translations = defaultdict(list)

    # Iterate over the JSON data to extract and store badge translations
    for entry in json_data:
        coords_dict = json.loads(entry['translations'])
        for badge, coords in coords_dict.items():
            if badge in CANDIDATES:
                x, y, z = coords  # Extracting x, y, and z coordinates
                badge_translations[badge].append({
                    'x': x[0],
                    'y': y[0],
                    'z': z[0]
                })

    # Initialize the Matplotlib figure for 2D heatmap using Z and X coordinates
    plt.figure(figsize=(12, 8))
    # Prepare data for heatmap
    all_z_coords = []
    all_x_coords = []
    for badge, coords_list in badge_translations.items():
        all_z_coords.extend([entry['z'] for entry in coords_list])
        all_x_coords.extend([entry['x'] for entry in coords_list])
    # Create the 2D histogram (heatmap)
    plt.hist2d(all_z_coords, all_x_coords, bins=(50, 50), cmap='inferno')
    # Add colorbar, labels, and title
    plt.colorbar(label='Frequency')
    plt.xlabel('Z Coordinate')
    plt.ylabel('X Coordinate')
    plt.title('2D Heatmap of Badge Positions (Z, X)')
    plt.savefig(os.path.join(visualization_dir, f"{bucket_name}_hm.png"), dpi=300)
    print(f"Saved 2D heatmap to {bucket_name}_hm.png")


def plot_physical_interaction_network(json_file_path, visualization_dir):
    """Plots a network graph representing physical interactions between badges."""
    json_data = read_json_file(json_file_path)
    df = convert_to_dataframe(json_data, ['graph', 'segment_start_time'])
    normalize_factor = len(df)  # Normalized by number of entries
    physical_interactions = calculate_physical_interactions(df, normalize_factor)
    session_name = os.path.basename(json_file_path).split('_')[1]

    # Create a directed graph
    G = nx.DiGraph()
    for src, targets in physical_interactions.items():
        for tgt, freq in targets.items():
            G.add_edge(src, tgt, weight=freq)

    # fixed_positions = {'0': (0, 0), '1': (0, 1), '2': (1, 1), '3': (1, 0)}  # specify positions here
    fixed_positions = None
    plt.figure(figsize=(10, 10))
    pos = nx.spring_layout(G, seed=42) if fixed_positions is None else fixed_positions

    # Calculate accumulative weights for each node
    accumulative_weights = {node: sum(weight for _, _, weight in G.edges(node, data='weight')) for node in G.nodes}

    # Node size parameters
    max_node_size = 1000  # Maximum node size
    min_node_size = 400  # Minimum node size

    # Normalize node sizes using a logarithmic scale
    node_sizes = {}
    for node, weight in accumulative_weights.items():
        node_sizes[node] = min_node_size + (weight - 0) / (2 - 0) * (
                max_node_size - min_node_size) if weight > 0 else min_node_size

    # Draw nodes with size based on accumulative weight
    nx.draw_networkx_nodes(G, pos, node_size=[node_sizes[node] for node in G.nodes], alpha=0.2)

    # Draw edges
    # Identify curved edges
    curved_edges = [edge for edge in G.edges() if G.has_edge(edge[1], edge[0])]
    straight_edges = list(set(G.edges()) - set(curved_edges))

    # Draw straight edges
    for (u, v) in straight_edges:
        wt = G[u][v]['weight']
        nx.draw_networkx_edges(G, pos, edgelist=[(u, v)], width=weight_to_width(wt, max_wt=1, max_width=10),
                               connectionstyle=f'arc3, rad = 0.2',
                               edge_color='moccasin', style='solid')

    # Draw curved edges with arc style
    for (u, v) in curved_edges:
        wt = G[u][v]['weight']
        nx.draw_networkx_edges(G, pos, edgelist=[(u, v)], width=weight_to_width(wt, max_wt=1, max_width=10),
                               connectionstyle=f'arc3, rad = 0.2',
                               edge_color='orange', style='dashed')

    # Draw edge labels using the custom function
    ax = plt.gca()
    arc_rad = 0.25
    edge_weights = nx.get_edge_attributes(G, 'weight')
    curved_edge_labels = {edge: round(edge_weights[edge], 4) for edge in curved_edges}
    straight_edge_labels = {edge: round(edge_weights[edge], 4) for edge in straight_edges}
    my_draw_networkx_edge_labels(G, pos, ax=ax, edge_labels=curved_edge_labels, rotate=True, rad=arc_rad)
    my_draw_networkx_edge_labels(G, pos, ax=ax, edge_labels=straight_edge_labels, rotate=True, rad=arc_rad)

    # Draw node labels
    nx.draw_networkx_labels(G, pos, font_size=7)

    plt.title(f'Normalized Physical Interaction Network ({session_name})')
    bucket_name = f"session_{os.path.basename(json_file_path).split('_')[1]}"
    plt.savefig(os.path.join(visualization_dir, f"{bucket_name}_npin.png"), dpi=300)


def calculate_physical_interactions(df, normalize_factor):
    """Calculates the frequencies of physical interactions between badges based on DataFrame data."""
    interaction_counts = defaultdict(lambda: defaultdict(int))

    for _, row in df.iterrows():
        graph_dict = row['graph']
        for src, targets in graph_dict.items():
            if src in CANDIDATES:
                for tgt in targets:
                    if tgt in CANDIDATES and tgt != src:
                        interaction_counts[src][str(tgt)] += 1 / normalize_factor

    return interaction_counts


def across_sessions_analysis_video(logs_dir, visualizations_dir):
    """Analyzes and visualizes various movement metrics across multiple sessions using saved log data."""
    session_data = []

    for bucket_dir_name in sorted(os.listdir(logs_dir)):
        bucket_dir_path = os.path.join(logs_dir, bucket_dir_name)
        if os.path.isdir(bucket_dir_path):
            session_name = bucket_dir_name.split('_')[1]
            # Initialize metrics with null values
            stm, nstm, srm, nsrm = None, None, None, None
            for filename in os.listdir(bucket_dir_path):
                if filename.endswith("_translations.json"):
                    translation_file = os.path.join(bucket_dir_path, filename)
                    _, stm, nstm = analyze_translations_log(translation_file)
                if filename.endswith("_rotations.json"):
                    rotation_file = os.path.join(bucket_dir_path, filename)
                    _, srm, nsrm = analyze_rotations_log(rotation_file)
            session_data.append((session_name, stm, nstm, srm, nsrm))

    # Unpack the session data for plotting
    session_names, stm_lst, nstm_lst, srm_lst, nsrm_lst = zip(*session_data)

    # Call plotting functions
    plot_across_sessions_analysis_video(session_names, stm_lst, nstm_lst, srm_lst, nsrm_lst, visualizations_dir)
    plot_across_sessions_analysis_video_interactive(session_names, stm_lst, nstm_lst, srm_lst, nsrm_lst,
                                                    visualizations_dir)


def plot_across_sessions_analysis_video(session_names, stm_lst, nstm_lst, srm_lst, nsrm_lst, save_dir):
    """Creates and saves matplotlib plots of various video analysis metrics across different sessions.

    Args:
        session_names (list): List of session names.
        stm_lst (list): List of Session Translate Movement values.
        nstm_lst (list): List of Normalized Session Translate Movement values.
        srm_lst (list): List of Session Rotate Movement values.
        nsrm_lst (list): List of Normalized Session Rotate Movement values.
        save_dir (str): Directory to save the plots.
    """
    # Create subplots
    fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(12, 12))

    # Plotting the first graph
    ax1.set_title("STM and NSTM")
    ax1.set_xlabel("Session")
    ax1.set_ylabel("Session Translate Movement (m)", color='tab:blue')
    ax1.plot(session_names, stm_lst, 'o-', color='tab:blue', label='STM')
    ax1.tick_params(axis="y", labelcolor="tab:blue")

    ax1_twin = ax1.twinx()
    ax1_twin.set_ylabel("Normalized Session Translate Movement (m/s)", color='tab:red')
    ax1_twin.plot(session_names, nstm_lst, 'x-', color='tab:red', label='NSTM')
    ax1_twin.tick_params(axis="y", labelcolor="tab:red")

    # Plotting the second graph
    ax2.set_title("SRM and NSRM")
    ax2.set_xlabel("Session")
    ax2.set_ylabel("Session Rotate Movement (°)", color='tab:green')
    ax2.plot(session_names, srm_lst, 's-', color='tab:green', label='SRM')
    ax2.tick_params(axis="y", labelcolor="tab:green")

    ax2_twin = ax2.twinx()
    ax2_twin.set_ylabel("Normalized Session Rotate Movement (°/s)", color='tab:purple')
    ax2_twin.plot(session_names, nsrm_lst, 'd-', color='tab:purple', label='NSRM')
    ax2_twin.tick_params(axis="y", labelcolor="tab:purple")

    # Show legends
    ax1.legend(loc='upper left')
    ax1_twin.legend(loc='upper right')
    ax2.legend(loc='upper left')
    ax2_twin.legend(loc='upper right')

    # Show plot
    fig.suptitle("Video Analysis Metrics Across Sessions")
    fig.tight_layout()

    plt.savefig(os.path.join(save_dir, 'video_analysis_across_sessions.png'), dpi=300)
    print("Saved video analysis metrics across sessions to video_analysis_across_sessions.png")


def plot_across_sessions_analysis_video_interactive(session_names, stm_lst, nstm_lst, srm_lst, nsrm_lst, save_dir):
    """Visualize video analysis metrics across sessions with pyecharts."""
    # Generate visualization page
    page = Page(page_title="video metrics analysis across sessions", layout=Page.SimplePageLayout)

    def format_list(data):
        return [f"{x:.4f}" if x is not None else '0' for x in data]

    stm_lst = format_list(stm_lst)
    nstm_lst = format_list(nstm_lst)
    srm_lst = format_list(srm_lst)
    nsrm_lst = format_list(nsrm_lst)
    yaxis_min_js = JsCode("function(value){return Math.round(value.min * 0.8 * 1000) / 1000;}")
    yaxis_max_js = JsCode("function(value){return Math.round(value.max * 1.2 * 1000) / 1000;}")

    # Create Line chart for STM and NSTM
    line1 = (
        Line(init_opts=opts.InitOpts(width="1500px", height="400px"))
        .add_xaxis(session_names)
        .add_yaxis("STM", stm_lst, is_smooth=False, yaxis_index=0)
        .add_yaxis("NSTM", nstm_lst, is_smooth=False, yaxis_index=1)
        .extend_axis(
            yaxis=opts.AxisOpts(
                name="Normalized Session Translate Movement(m/s)",
                min_=yaxis_min_js,
                max_=yaxis_max_js,
                type_="value",
                position="right",
            )
        )
        .set_global_opts(
            title_opts=opts.TitleOpts(title="STM and NSTM", pos_top="0%", pos_left="center"),
            legend_opts=opts.LegendOpts(
                pos_top="7%",
                item_width=20,
                item_height=10,
                textstyle_opts=opts.TextStyleOpts(font_size=10)
            ),
            yaxis_opts=opts.AxisOpts(
                name="Session Translate Movement(m)",
                min_=yaxis_min_js,
                max_=yaxis_max_js,
            ),
            xaxis_opts=opts.AxisOpts(name="Session", name_location="middle", name_gap=30)
        )
    )

    # Create Line chart for SRM and NSRM
    line2 = (
        Line(init_opts=opts.InitOpts(width="1500px", height="400px"))
        .add_xaxis(session_names)
        .add_yaxis("SRM", srm_lst, is_smooth=False, yaxis_index=0)
        .add_yaxis("NSRM", nsrm_lst, is_smooth=False, yaxis_index=1)
        .extend_axis(
            yaxis=opts.AxisOpts(
                name="Normalized Session Rotate Movement(°/s)",
                min_=yaxis_min_js,
                max_=yaxis_max_js,
                type_="value",
                position="right",
            )
        )
        .set_global_opts(
            title_opts=opts.TitleOpts(title="SRM and NSRM", pos_top="0%", pos_left="center"),
            legend_opts=opts.LegendOpts(
                pos_top="7%",
                item_width=20,
                item_height=10,
                textstyle_opts=opts.TextStyleOpts(font_size=10)
            ),
            yaxis_opts=opts.AxisOpts(
                name="Session Rotate Movement(°)",
                min_=yaxis_min_js,
                max_=yaxis_max_js,
            ),
            xaxis_opts=opts.AxisOpts(name="Session", name_location="middle", name_gap=30)
        )
    )

    page.add(line1)
    page.add(line2)
    page.render(os.path.join(save_dir, f'video_analysis_across_sessions.html'))
    print("Saved interactive video analysis metrics across sessions to video_analysis_across_sessions.html")


def analyze_translations_log(json_path):
    """Calculates translation-based movement levels using Euclidean distances between translation vectors."""
    json_data = read_json_file(json_path)
    df = convert_to_dataframe(json_data, ['segment_start_time', 'translations'])
    session_duration = len(df) * 1
    badges_movement = {}

    for index, row in df.iterrows():
        translations = row['translations']
        for badge_id, coordinates in translations.items():
            coordinates = np.array(coordinates).flatten()
            if badge_id not in badges_movement:
                badges_movement[badge_id] = {'previous_coordinates': None, 'total_movement': 0}
            if badges_movement[badge_id]['previous_coordinates'] is not None:
                distance = euclidean(badges_movement[badge_id]['previous_coordinates'], coordinates)
                badges_movement[badge_id]['total_movement'] += distance
            badges_movement[badge_id]['previous_coordinates'] = coordinates

    individual_badge_movement = {badge_id: data['total_movement'] for badge_id, data in badges_movement.items()}
    total_session_movement = sum(individual_badge_movement.values())
    normalized_session_movement = total_session_movement / session_duration
    return individual_badge_movement, total_session_movement, normalized_session_movement


def analyze_rotations_log(json_path):
    """Computes rotation-based movement levels using degree changes between rotation matrices."""
    json_data = read_json_file(json_path)
    df = convert_to_dataframe(json_data, ['segment_start_time', 'rotations'])
    session_duration = len(df) * 1  # Assuming each row represents 1 unit of time
    badges_movement = {}

    for index, row in df.iterrows():
        rotations = row['rotations']
        for badge_id, rotation_matrix in rotations.items():
            if badge_id not in badges_movement:
                badges_movement[badge_id] = {'previous_rotation': None, 'total_movement': 0}
            if badges_movement[badge_id]['previous_rotation'] is not None:
                # Calculate the rotation matrix that represents the rotation from previous to current
                R_diff = np.dot(np.linalg.inv(np.array(badges_movement[badge_id]['previous_rotation'])),
                                np.array(rotation_matrix))
                # Calculate the degree of rotation
                degree_of_rotation = calculate_degree_of_rotation(R_diff)
                badges_movement[badge_id]['total_movement'] += degree_of_rotation
            badges_movement[badge_id]['previous_rotation'] = np.array(rotation_matrix)

    individual_badge_movement = {badge_id: data['total_movement'] for badge_id, data in badges_movement.items()}
    total_session_movement = sum(individual_badge_movement.values())
    normalized_session_movement = total_session_movement / session_duration
    return individual_badge_movement, total_session_movement, normalized_session_movement


def calculate_degree_of_rotation(R):
    """Calculate the degree of rotation from a rotation matrix."""
    trace_value = np.trace(R)
    angle_rad = np.arccos((trace_value - 1) / 2)
    angle_deg = np.degrees(angle_rad)
    return angle_deg



import os

import cv2
import numpy as np
from pupil_apriltags import Detector


def detect_apriltags(image_path, tag_family='tag36h11', render=True, show=True, save=False):
    """
    Detect the apriltags in image and return the positions of the tags.

    Args:
        image_path: Path to the image file
        tag_family: Tag family to detect
        render: Render the detected tags on the image or not
        show: Show the detected image or not
        save: Save the detected image or not
    """
    tag_pos = {}

    # Load image
    image = cv2.imread(image_path)
    if image is None:
        print("Failed to load image")
        return

    # Image resolution
    height, width, _ = image.shape
    print(f"Image resolution: {width}x{height} (Width x Height)")

    # Detect tags in the image
    detector = Detector(families=tag_family, nthreads=4)
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    tags = detector.detect(gray)

    # Iterate over detected tags
    for tag in tags:
        # Calculate the center of the tag
        center = np.mean(tag.corners, axis=0)
        corners = np.int32(tag.corners)

        # Draw the tag ID with a background and border
        if render:
            tag_position = (int(corners[0][0]), int(corners[0][1]) - 10)  # Adjust position above the tag
            text = f"{tag.tag_id}"
            font = cv2.FONT_HERSHEY_SIMPLEX
            font_scale = 5
            font_thickness = 14
            text_size, _ = cv2.getTextSize(text, font, font_scale, font_thickness)
            text_x, text_y = tag_position
            # Draw black border
            cv2.rectangle(image, (text_x - 20, text_y - text_size[1] - 20), (text_x + text_size[0] + 10, text_y + 30),
                          (0, 0, 0), -1)
            # Draw white background
            cv2.rectangle(image, (text_x - 10, text_y - text_size[1] - 10), (text_x + text_size[0], text_y + 20),
                          (255, 255, 255), -1)
            # Put black text
            cv2.putText(image, text, (text_x, text_y), font, font_scale, (0, 0, 0), font_thickness)

        # Print the center position
        x = float(f'{center[0] / width:.4f}')
        y = float(f'{center[1] / height:.4f}')
        print(f"Person ID {tag.tag_id} center position: [{x}, {y}]")
        tag_pos[tag.tag_id] = [x, y]

    # Show the result
    if show:
        cv2.imshow('AprilTag Detection', image)
        cv2.waitKey(0)
        cv2.destroyAllWindows()

    # Save detected image
    if save:
        dirname, filename = os.path.split(image_path)
        name, ext = os.path.splitext(filename)
        save_path = os.path.join(dirname, f"{name}_detected{ext}")
        cv2.imwrite(save_path, image)
        print(f"Detected image saved as {save_path}")

    return tag_pos



import gc
import inspect
import json
import logging
import os
import threading
import time

from openmmla_vision.utils.analyze_utils import session_analysis_video
from openmmla_vision.utils.influx_client import InfluxDBClientWrapper
from openmmla_vision.utils.input_utils import get_function_synchronizer, get_bucket_name
from openmmla_vision.utils.logger_utils import get_logger
from openmmla_vision.utils.mqtt_client import MQTTClientWrapper
from openmmla_vision.utils.redis_client import RedisClientWrapper
from openmmla_vision.utils.relations_utils import is_tag_looking_at_another_2d
from openmmla_vision.utils.thread_utils import RaisingThread
from openmmla_vision.utils.transform_utils import transform_point, transform_rotation


class Synchronizer:
    """Synchronizer class for synchronizing detection results from multiple cameras and uploading to InfluxDB"""
    logger = get_logger('synchronizer')

    def __init__(self, project_dir: str = None, config_path: str = None):
        """Initialize the synchronizer.

        Args:
            project_dir: path to the project directory
            config_path: path to the configuration file
        """
        # Set the project root directory
        if project_dir is None:
            caller_frame = inspect.stack()[1]
            caller_module = inspect.getmodule(caller_frame[0])
            if caller_module is None:
                self.project_dir = os.getcwd()
            else:
                self.project_dir = os.path.dirname(os.path.abspath(caller_module.__file__))
        else:
            self.project_dir = project_dir

        # Determine the configuration path
        if config_path:
            if os.path.isabs(config_path):
                self.config_path = config_path
            else:
                self.config_path = os.path.join(self.project_dir, config_path)
        else:
            self.config_path = os.path.join(self.project_dir, 'conf/video_base.ini')

        # Check if the configuration file exists
        if not os.path.exists(self.config_path):
            raise FileNotFoundError(f"Configuration file not found at {self.config_path}")

        self.logger_dir = os.path.join(self.project_dir, 'logger')
        self.camera_sync_dir = os.path.join(self.project_dir, 'camera_sync')
        os.makedirs(self.logger_dir, exist_ok=True)
        os.makedirs(self.camera_sync_dir, exist_ok=True)

        self.main_camera_id = None
        self.transform_matrices_dict = None
        self.bucket_name = None
        self.graph_dict = None
        self.location_dict = None
        self.alive = False
        self.threads = []
        self.lock = threading.Lock()
        self.stop_event = threading.Event()

        self.influx_client = InfluxDBClientWrapper(self.config_path)
        self.mqtt_client = MQTTClientWrapper(self.config_path)
        self.redis_client = RedisClientWrapper(self.config_path)

    def run(self):
        print('\033]0;Video Synchronizer\007')
        func_map = {1: self._start_synchronizing, 2: self._set_main_camera}

        while True:
            try:
                select_fun = get_function_synchronizer()
                if select_fun == 0:
                    self.logger.info("Exiting video synchronizer...")
                    break
                func_map.get(select_fun, lambda: print("Invalid option."))()
            except (Exception, KeyboardInterrupt) as e:
                self.logger.warning(
                    f"During running the synchronizer, catch: {'KeyboardInterrupt' if isinstance(e, KeyboardInterrupt) else e}, Come back to the main menu.",
                    exc_info=True)

    def _start_synchronizing(self):
        if self.transform_matrices_dict is None:
            self.logger.warning("Main camera id or transformation matrices not set, please set them first.")
            return self._set_main_camera()

        self.graph_dict = {}
        self.location_dict = {}
        self.bucket_name = get_bucket_name(self.influx_client)
        self.logger = get_logger(f'synchronizer-{self.bucket_name}',
                                 os.path.join(self.logger_dir, f'{self.bucket_name}_synchronizer.log'),
                                 level=logging.DEBUG, console_level=logging.INFO, file_level=logging.DEBUG)
        exception_occurred = None

        self._listen_for_start_signal()

        # Reinitialize MQTT client with new topic and on_message callback
        self.mqtt_client.reinitialise(on_message=self._handle_base_result, topics=f'{self.bucket_name}/video')
        self.mqtt_client.loop_start()

        # Create threads
        self._create_thread(self._listen_for_stop_signal)
        self._create_thread(self._upload_data)

        # Start threads
        try:
            self._start_threads()
            self._join_threads()
        except (Exception, KeyboardInterrupt) as e:
            self.logger.warning(
                f"During voice recognition, catch: {'KeyboardInterrupt' if isinstance(e, KeyboardInterrupt) else e}",
                exc_info=True)
            exception_occurred = e
        finally:
            self._synchronization_handler(exception_occurred)

    def _set_main_camera(self):
        self.transform_matrices_dict = self._load_transform_matrices()
        if self.transform_matrices_dict is None:
            self.logger.warning("No transformation matrices found, please check your main camera id or do the camera "
                                "sync first.")

    def _synchronization_handler(self, e):
        """Handle exceptions and stop all threads."""
        if e:
            self._stop_threads()
        else:
            self.logger.info("All threads stopped properly.")

        self.mqtt_client.loop_stop()
        session_analysis_video(self.project_dir, self.bucket_name, self.influx_client)
        self._clean_up()

    def _create_thread(self, target, *args):
        """Create a new thread and add it to the thread list.

        :param target: the target function to run in the thread
        """
        t = RaisingThread(target=target, args=args)
        self.threads.append(t)

    def _start_threads(self):
        """Start all threads."""
        self.stop_event.clear()
        for t in self.threads:
            t.start()

    def _join_threads(self):
        """Wait for all threads to finish."""
        for t in self.threads:
            t.join()

    def _stop_threads(self):
        """Stop all threads and free memory."""
        self.stop_event.set()
        for t in self.threads:
            if threading.current_thread() != t:
                try:
                    t.join(timeout=5)
                except Exception as e:
                    self.logger.warning("During the thread stopping, catch exception %s", e, exc_info=True)
        self.threads.clear()

    def _clean_up(self):
        """Free memory by resetting dictionaries."""
        self.bucket_name = None
        self.graph_dict = None
        self.location_dict = None
        gc.collect()

    def _listen_for_start_signal(self):
        """Listen on the redis bucket control channel for the START signal."""
        p = self.redis_client.subscribe(f"{self.bucket_name}/control")
        self.logger.info("Wait for START signal...")
        while True:
            message = p.get_message()
            if message and message['data'] == b'START':
                self.logger.info("Received START signal, start synchronizing...")
                break

    def _listen_for_stop_signal(self):
        """Listen on the redis bucket control channel for the STOP signal."""
        p = self.redis_client.subscribe(f"{self.bucket_name}/control")
        self.logger.info("Listening for STOP command...")

        while not self.stop_event.is_set():
            message = p.get_message()
            if message and message['data'] == b'STOP':
                self.logger.info("Received STOP signal, stop synchronizing...")
                self.stop_event.set()

    def _handle_base_result(self, client, userdata, msg):
        with self.lock:
            if not self.stop_event.is_set():
                self.alive = True
                message = json.loads(msg.payload)
                sender_id = message["sender_id"]

                if sender_id.isnumeric():  # msg from nicla vision camera
                    self.graph_dict.setdefault(sender_id, []).extend(message['detected_tags'])
                else:  # msg from base camera
                    tags = message["tags"]
                    tag_relations = message["tag_relations"]

                    if sender_id != self.main_camera_id:  # convert to main coordinates
                        R = self.transform_matrices_dict[sender_id]['R']
                        T = self.transform_matrices_dict[sender_id]['T']
                        for tag_id, tag_data in tags.items():
                            main_rotation = transform_rotation(R, tag_data[0])
                            main_translation = transform_point(tag_data[1], R, T)
                            self.location_dict[tag_id] = [main_rotation, main_translation]
                    else:
                        self.location_dict.update(tags)

                    # Store tag relations into graph
                    for tag_id, look_at_tags in tag_relations.items():
                        self.graph_dict.setdefault(tag_id, []).extend(look_at_tags)

                    # Detect tag relations again under main camera's coordinate system
                    for tag_id, tag_data in self.location_dict.items():
                        self.graph_dict.setdefault(tag_id, [])
                        for target_id, target_data in self.location_dict.items():
                            if target_id != tag_id and target_id not in self.graph_dict[tag_id]:
                                if is_tag_looking_at_another_2d(tag_data, target_data, angle_threshold=-0.7,
                                                                distance_threshold=0.7):
                                    self.graph_dict[tag_id].append(target_id)

    def _upload_data(self):
        """Upload data to InfluxDB for every 1 second"""
        while not self.stop_event.is_set():
            with self.lock:
                if self.alive:
                    segment_start_time = int(time.time()) - 1.0
                    rotations_dict = {}
                    translations_dict = {}
                    for tag_id, (rotation, translation) in self.location_dict.items():
                        rotations_dict[tag_id] = rotation
                        translations_dict[tag_id] = translation

                    # Prepare and upload the data for badge translations,rotations and relations
                    self._write_to_influx("badge translations", {
                        "segment_start_time": segment_start_time,
                        "translations": json.dumps(translations_dict),
                    })
                    self._write_to_influx("badge rotations", {
                        "segment_start_time": segment_start_time,
                        "rotations": json.dumps(rotations_dict),
                    })
                    self._write_to_influx("badge relations", {
                        "segment_start_time": segment_start_time,
                        "graph": json.dumps(self.graph_dict),
                    })

                # Reset for next cycle
                self.graph_dict.clear()
                self.location_dict.clear()
                self.alive = False

            # Schedule the next upload outside the lock to avoid potential deadlocks
            time.sleep(1)

    def _write_to_influx(self, measurement, fields):
        data = {
            "measurement": measurement,
            "fields": fields,
        }
        # self.logger.info(f"{data}")
        self.influx_client.write(self.bucket_name, record=data)

    def _load_transform_matrices(self):
        transformation_choices = [d for d in os.listdir(self.camera_sync_dir) if
                                  d.startswith('transformation_matrices_')]
        for idx, choice in enumerate(transformation_choices):
            print(f"{idx}: {choice}")

        if not transformation_choices:
            return None

        while True:
            try:
                selection = int(input("Choose your main transformation matrices with number: "))
                if not 0 <= selection < len(transformation_choices):
                    self.logger.warning("Invalid selection. Please choose a valid number.")
                else:
                    chosen_transformation = transformation_choices[selection]
                    self.main_camera_id = chosen_transformation.removesuffix('.json').split('_')[-1]
                    self.logger.info(f"Main camera id: {self.main_camera_id} has been selected.")
                    break
            except ValueError:
                self.logger.warning("Please enter a valid number.")

        with open(os.path.join(self.camera_sync_dir, chosen_transformation), 'r') as file:
            return json.load(file)



import distutils.util


def print_arguments(args):
    print("-----------  Configuration Arguments -----------")
    for arg, value in sorted(vars(args).items()):
        print("%s: %s" % (arg, value))


def add_arguments(argname, type, default, help, argparser, shortname=None, **kwargs):
    type = distutils.util.strtobool if type == bool else type
    args = ["--" + argname]
    if shortname:
        args.insert(0, shortname)
    argparser.add_argument(*args,
                           default=default,
                           type=type,
                           help=help + ' Default: %(default)s.',
                           **kwargs)



import configparser

import redis


class RedisClientWrapper(redis.Redis):
    """Extended Redis client that loads configuration from a file and adds custom functionalities."""

    def __init__(self, config_path):
        """Initialize Redis client with configurations"""
        config = configparser.ConfigParser()
        config.read(config_path)

        super().__init__(
            host=config['Redis']['redis_host'],
            port=int(config['Redis']['redis_port']),
            health_check_interval=10,
            socket_timeout=10,
            socket_keepalive=True,
            socket_connect_timeout=10,
            retry_on_timeout=True
        )

    def subscribe(self, topic):
        """Subscribe to a topic and return a pubsub object"""
        p = self.pubsub()
        p.subscribe(topic)
        return p



import numpy as np


def is_tag_looking_at_another(tag1, tag2, angle_threshold, distance_threshold):
    """Check if tag1 is looking at tag2 considering their orientation and distance."""
    if isinstance(tag1, list):
        distance = np.linalg.norm(np.array(tag1[1]).ravel() - np.array(tag2[1]).ravel())
    else:
        distance = np.linalg.norm(tag1.pose_t.ravel() - tag2.pose_t.ravel())
    if distance > distance_threshold:
        return False

    normal1, _ = get_outward_normal(tag1)
    vector_from_tag2_to_tag1 = unit_vector_from_tag_to_tag(tag2, tag1)
    score = np.dot(normal1, vector_from_tag2_to_tag1)

    return score < angle_threshold


def get_outward_normal(tag):
    """Get the normal vector of the tag which perpendicular to the surface"""
    if isinstance(tag, list):  # for tag data [rotations, translations]
        normal = np.dot(np.array(tag[0]), np.array([0, 0, 1])).ravel()
        normal = -normal
        tag = None
    # for tag object
    else:
        normal = np.dot(tag.pose_R, np.array([0, 0, 1])).ravel()
        if np.dot(normal, tag.pose_t.ravel()) < 0:
            tag.pose_R[:, 2] = -tag.pose_R[:, 2]  # Flip the sign of the third column of the rotation matrix
        else:
            normal = -normal

    return normal, tag


def unit_vector_from_tag_to_tag(source_tag, target_tag):
    """Return the unit vector pointing from the source tag to the target tag"""
    if isinstance(source_tag, list):
        direction = np.array(target_tag[1]).ravel() - np.array(source_tag[1]).ravel()
    else:
        direction = target_tag.pose_t.ravel() - source_tag.pose_t.ravel()

    return direction / np.linalg.norm(direction)


def is_tag_looking_at_another_2d(tag1, tag2, angle_threshold, distance_threshold):
    """Check if tag1 is looking at tag2 considering their orientation and distance on the x-z plane."""
    if isinstance(tag1, list):
        distance = np.linalg.norm(np.array([tag1[1][0], tag1[1][2]]) - np.array([tag2[1][0], tag2[1][2]]))
    else:
        distance = np.linalg.norm(
            np.array([tag1.pose_t[0], tag1.pose_t[2]]) - np.array([tag2.pose_t[0], tag2.pose_t[2]]))
    # print(f"distance: {distance}")

    if distance > distance_threshold:
        return False

    normal1, _ = get_outward_normal_2d(tag1)
    vector_from_tag2_to_tag1 = unit_vector_from_tag_to_tag_2d(tag2, tag1)
    score = np.dot(normal1, vector_from_tag2_to_tag1)

    return score < angle_threshold


def get_outward_normal_2d(tag):
    """Get the normal vector of the tag which is perpendicular to the surface and project it onto the x-z plane."""
    if isinstance(tag, list):  # for tag data [rotations, translations] in synchronizer
        normal = np.dot(np.array(tag[0]), np.array([0, 0, 1])).ravel()
        normal = -normal
        tag = None
    else:  # for tag object in video base
        normal = np.dot(tag.pose_R, np.array([0, 0, 1])).ravel()
        if np.dot(normal, tag.pose_t.ravel()) < 0:
            tag.pose_R[:, 2] = -tag.pose_R[:, 2]  # Flip the sign of the third column of the rotation matrix
        else:
            normal = -normal

    normal_xz = np.array([normal[0], 0, normal[2]])  # Project onto x-z plane
    normal_xz /= np.linalg.norm(normal_xz)  # Normalize

    return normal_xz, tag


def unit_vector_from_tag_to_tag_2d(source_tag, target_tag):
    """Return the unit vector pointing from the source tag to the target tag on the x-z plane"""
    if isinstance(source_tag, list):
        direction = np.array(target_tag[1]).ravel() - np.array(source_tag[1]).ravel()
    else:
        direction = target_tag.pose_t.ravel() - source_tag.pose_t.ravel()
    direction_xz = np.array([direction[0], 0, direction[2]])  # Project onto x-z plane

    return direction_xz / np.linalg.norm(direction_xz)



import inspect
import json
import os
import platform
import time
from multiprocessing import Process

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from matplotlib.animation import FuncAnimation
from numpy.linalg import norm

from openmmla_vision.utils.influx_client import InfluxDBClientWrapper
from openmmla_vision.utils.input_utils import get_bucket_name, get_function_visualizer
from openmmla_vision.utils.logger_utils import get_logger
from openmmla_vision.utils.redis_client import RedisClientWrapper

# https://stackoverflow.com/questions/44144584/typeerror-cant-pickle-thread-lock-objects/78013322#78013322
# Add 'export OBJC_DISABLE_INITIALIZE_FORK_SAFETY=YES' to ~/.zshrc for multiprocess fork
if platform.system() != "Linux":
    from multiprocessing import set_start_method

    set_start_method("fork")


class Visualizer:
    """Real Time Visualization class for visualizing the real-time badge relations and positions"""
    logger = get_logger('visualizer')

    def __init__(self, project_dir: str = None, config_path: str = None, store: bool = False):
        """Initialize the visualizer.

        Args:
            project_dir: path to the project directory
            config_path: path to the configuration file
            store: whether to store the visualization images in the local directory
        """
        # Set the project root directory
        if project_dir is None:
            caller_frame = inspect.stack()[1]
            caller_module = inspect.getmodule(caller_frame[0])
            if caller_module is None:
                self.project_dir = os.getcwd()
            else:
                self.project_dir = os.path.dirname(os.path.abspath(caller_module.__file__))
        else:
            self.project_dir = project_dir

        # Determine the configuration path
        if config_path:
            if os.path.isabs(config_path):
                self.config_path = config_path
            else:
                self.config_path = os.path.join(self.project_dir, config_path)
        else:
            self.config_path = os.path.join(self.project_dir, 'conf/video_base.ini')

        # Check if the configuration file exists
        if not os.path.exists(self.config_path):
            raise FileNotFoundError(f"Configuration file not found at {self.config_path}")

        self.store = store

        self.visualizations_dir = os.path.join(self.project_dir, 'visualizations')
        os.makedirs(self.visualizations_dir, exist_ok=True)

        self.bucket_name = None
        self.processes = []

        self.redis_client = RedisClientWrapper(self.config_path)
        self.influx_client_main = InfluxDBClientWrapper(self.config_path)

    def run(self):
        print('\033]0;Video Visualizer\007')
        func_map = {1: self._start_visualizing}

        while True:
            try:
                select_fun = get_function_visualizer()
                if select_fun == 0:
                    self.logger.info("Exiting video visualizer...")
                    break
                func_map.get(select_fun, lambda: print("Invalid option."))()
            except (Exception, KeyboardInterrupt) as e:
                self.logger.warning(
                    f"During running the visualizer, catch: {'KeyboardInterrupt' if isinstance(e, KeyboardInterrupt) else e}, Come back to the main menu.",
                    exc_info=True)

    def _start_visualizing(self):
        self.bucket_name = get_bucket_name(self.influx_client_main)

        if self.store:
            dir_path = os.path.join(self.visualizations_dir, f'{self.bucket_name}/real-time')
            os.makedirs(dir_path, exist_ok=True)

        self._listen_for_start_signal()
        self.create_process(self._start_2d_plot)
        self.create_process(self._start_3d_plot)

        try:
            self.start_processes()
            self.join_processes()
        except (Exception, KeyboardInterrupt) as e:
            self.logger.warning("%s, Come back to the main menu.", e, exc_info=True)
        finally:
            self.stop_processes()
            self.bucket_name = None

    def create_process(self, target):
        p = Process(target=target, daemon=True)
        self.processes.append(p)

    def start_processes(self):
        for p in self.processes:
            p.start()

    def join_processes(self):
        for p in self.processes:
            p.join()

    def stop_processes(self):
        for p in self.processes:
            if p.is_alive():
                p.terminate()  # Send termination request
                p.join()  # Wait for the process to finish
        self.processes.clear()

    def _listen_for_start_signal(self):
        p = self.redis_client.subscribe(f"{self.bucket_name}/control")
        self.logger.info("Wait for START signal...")

        while True:
            message = p.get_message()
            if message and message['data'] == b'START':
                self.logger.info("Received START signal, start visualizing...")
                break

    def _start_2d_plot(self):
        influx_client = InfluxDBClientWrapper(self.config_path)
        fig = plt.figure()
        ani = FuncAnimation(fig, self._animate, fargs=(influx_client,), interval=50, cache_frame_data=False)
        plt.show()

    def _start_3d_plot(self):
        influx_client = InfluxDBClientWrapper(self.config_path)
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        ani = FuncAnimation(fig, self._animate_3d, fargs=(fig, ax, influx_client,),
                            interval=50, cache_frame_data=False)
        plt.show()

    def _animate(self, i, influx_client):
        plt.cla()

        # Get node relations and pos
        graph_dict, segment_time = self._get_node_relations(influx_client)
        if graph_dict is None:
            return
        pos = self._get_node_positions(influx_client, segment_time=segment_time, dimension='2d')
        G = self._build_graph(graph_dict, pos)

        options = {
            "font_size": 15,
            "node_size": 1000,
            "node_color": "white",
            "edgecolors": "black",
            "linewidths": 3,
            "width": 3,
        }
        nx.draw_networkx(G, pos, **options)

        # Set margins for the axes so that nodes aren't clipped
        ax = plt.gca()
        ax.margins(0.20)
        plt.axis("off")
        if self.store:
            plt.savefig(
                os.path.join(self.visualizations_dir, f'{self.bucket_name}/real-time/image_{segment_time}_2d.png'))

    def _animate_3d(self, i, fig, ax, influx_client):
        plt.cla()

        # Get node relations and pos
        graph_dict, segment_time = self._get_node_relations(influx_client)
        if graph_dict is None:
            return
        pos_3d = self._get_node_positions(segment_time=segment_time, dimension='3d', influx_client=influx_client)
        G = self._build_graph(graph_dict, pos_3d)

        # Draw the 3D graph
        for node, coordinates in pos_3d.items():
            camera_x, camera_y, camera_z = coordinates
            ax.scatter(camera_x, camera_z, camera_y, s=200, c='white', edgecolors='green')
            ax.text(camera_x, camera_z, camera_y, node, fontsize=6, color='green', ha='center', va='center', zorder=40)

        # Draw directed edges as arrows
        for edge in G.edges():
            x1, y1, z1 = pos_3d[edge[0]]
            x2, y2, z2 = pos_3d[edge[1]]
            self.draw_arrow(ax, x1, z1, y1, x2, z2, y2)

        ax.set_xlabel('x')
        ax.set_ylabel('z')
        ax.set_zlabel('y')
        ax.view_init(elev=20., azim=30)  # adjust the viewing angle for better visualization
        if self.store:
            plt.savefig(
                os.path.join(self.visualizations_dir, f'{self.bucket_name}/real-time/image_{segment_time}_3d.png'))

    def _build_graph(self, graph_dict, pos):
        G = nx.DiGraph()
        G.add_node('B')
        # Add nodes and edges based on the graph_dict
        for badge_id, detected_tags in graph_dict.items():
            if badge_id not in G:
                G.add_node(badge_id)
            for tag_id in detected_tags:
                if str(tag_id) not in G:
                    G.add_node(str(tag_id))
                G.add_edge(badge_id, str(tag_id))

        # Remove missing nodes
        missing_nodes = [node for node in G.nodes() if node not in pos]
        if missing_nodes:
            self.logger.warning("Missing positions for nodes: %s", missing_nodes)
        for node in missing_nodes:
            G.remove_node(node)
        return G

    def _get_node_relations(self, influx_client):
        start_time = int(time.time()) - 20
        query = f"""from(bucket: "{self.bucket_name}")
                    |> range(start: {start_time})
                    |> last()
                    |> filter(fn: (r) => r._measurement == "badge relations")
                    |> pivot(rowKey: ["_time"], columnKey: ["_field"], valueColumn: "_value")
                    """
        tables = influx_client.query(query)
        data = json.loads(tables.to_json(indent=5))
        if not data:
            print("No data found for the specified bucket.")
            return None, None
        graph_dict_str = data[0]["graph"]
        graph_dict = json.loads(graph_dict_str)
        segment_time = data[0]["segment_start_time"]
        return graph_dict, segment_time

    def _get_node_positions(self, influx_client, segment_time, dimension='2d'):
        start_time = int(segment_time) - 20
        query = f"""from(bucket: "{self.bucket_name}")
                   |> range(start: {start_time})
                   |> filter(fn: (r) => r._measurement == "badge translations")
                   |> pivot(rowKey: ["_time"], columnKey: ["_field"], valueColumn: "_value")
                   |> filter(fn: (r) => r.segment_start_time == {segment_time})
                  """
        tables = influx_client.query(query)
        data = json.loads(tables.to_json(indent=5))
        translate_dict = json.loads(data[0]["translations"])

        positions = {'B': (0, 0)} if dimension == '2d' else {'B': (0, 0, 0)}
        for badge_id, translation in translate_dict.items():
            if dimension == '2d':
                x = translation[0][0]
                z = translation[2][0]
                positions[badge_id] = (z, -x)
            else:
                x = translation[0][0]
                y = translation[1][0]
                z = translation[2][0]
                positions[badge_id] = (x, -y, z)
        return positions

    @staticmethod
    def draw_arrow(ax, x1, y1, z1, x2, y2, z2, node_radius=0.04):
        arrow_vector = np.array([x2 - x1, y2 - y1, z2 - z1])
        arrow_unit_vector = arrow_vector / norm(arrow_vector)
        start_point = np.array([x1, y1, z1]) + node_radius * arrow_unit_vector
        end_point = np.array([x2, y2, z2]) - node_radius * arrow_unit_vector
        ax.plot([start_point[0], end_point[0]], [start_point[1], end_point[1]], [start_point[2], end_point[2]],
                c='green', linewidth=0.5, zorder=4)
        ax.scatter([end_point[0]], [end_point[1]], [end_point[2]], c='red', s=10, marker='.', zorder=10)



import os
import subprocess
from pathlib import Path

from setuptools import setup
from setuptools.command.install import install

NAME = 'openmmla-vision'
DESCRIPTION = 'Vision module for the OpenMMLA platform.'

URL = 'https://github.com/ucph-ccs/mbox-audio'
EMAIL = 'lizaibeim@gmail.com'
AUTHOR = 'Zaibei Li'
REQUIRES_PYTHON = '>=3.9.0'
VERSION = "0.1.0"

HERE = Path(__file__).parent

try:
    with open(HERE / "README.md", encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


def execute_script(script_name):
    script_path = os.path.join(os.path.dirname(__file__), script_name)
    if os.path.exists(script_path):
        print(f"Executing {script_name} script...")
        try:
            subprocess.check_call(['bash', script_path])
        except subprocess.CalledProcessError as e:
            print(f"Failed to execute {script_name}: {str(e)}")
    else:
        print(f"{script_name} script not found.")


class CustomInstallCommand(install):
    def run(self):
        install.run(self)
        execute_script('reset.sh')


# Package and dependency configuration
install_requires = [
    'influxdb-client==1.44.0',
    'matplotlib==3.9.1',
    'opencv-python==4.10.0.84',
    'paho-mqtt==2.1.0',
    'pandas==2.2.2',
    'pupil_apriltags==1.0.4.post10',
    'redis==5.0.7',
    'scipy==1.14.0',
    'networkx==3.3',
    'pyecharts==2.0.6',
]

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=['openmmla_vision', 'openmmla_vision/services', 'openmmla_vision/utils', 'openmmla_vision/bases'],
    install_requires=install_requires,
    include_package_data=True,
    license='MIT License',
    classifiers=[
        'Topic :: Multimedia :: Video :: Capture',
        'Topic :: Education :: Computer Aided Instruction (CAI)',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: MIT License',
    ],
    cmdclass={
        'install': CustomInstallCommand,
    }
)



import json
import math

import numpy as np
from scipy.spatial.transform import Rotation


def transform_rotation(R, R_alt):
    """Convert the rotation matrix from the alternative coordinate system to the main coordinate system.

    Args:
        R (3x3 list) : Rotation matrix to convert from alternative to main coordinate system
        R_alt (3x3 list): Rotation matrix in the alternative coordinate system

    Returns:
        R_transformed (3x3 list): Transformed rotation matrix in the main coordinate system
    """
    R = np.array(R)
    R_alt = np.array(R_alt)
    R_transformed = np.dot(R, R_alt)

    return R_transformed.tolist()


def transform_point(point, R, T):
    """Transform a point from the alternative coordinate system to the main coordinate system.

    Args:
        point (1x3 list): Point in the alternative coordinate system
        R (3x3 list): Rotation matrix to convert from alternative to main
        T (3x1 list): Translation vector to convert from alternative to main

    Returns:
        transformed_point (1x3 list): Transformed point in the main coordinate system
    """
    point = np.array(point)
    R = np.array(R)
    T = np.array(T)
    transformed_point = np.dot(R, point) + T

    return transformed_point.tolist()


def distance_between_points(point1, point2):
    """Calculate the Euclidean distance between two points.

    Args:
        point1 (1x3 list): First point
        point2 (1x3 list): Second point

    Returns:
        distance (float): Euclidean distance between point1 and point2
    """
    point1 = np.array(point1)
    point2 = np.array(point2)
    distance = np.linalg.norm(point1 - point2)

    return distance


def distance_between_rotations(P, Q):
    """The distance between rotations represented by rotation matrices and is the angle of the difference rotation
    represented by the rotation matrix R=PQ*.

    ref: http://www.boris-belousov.net/2016/12/01/quat-dist/#using-rotation-matrices

    Args:
        P (3x3 list): Rotation matrix P
        Q (3x3 list): Rotation matrix Q

    Returns:
        distance (float): The distance between the rotations P and Q in degrees
    """
    P = np.array(P)
    Q = np.array(Q)
    R = np.dot(P, Q.T)
    value = np.clip((np.trace(R) - 1) / 2, -1, 1)
    theta = math.acos(value)

    return math.degrees(theta)


def direct_transform_matrices(R_main, T_main, R_alt, T_alt):
    """Find transform matrices that converts alternative rotation and translation to the main coordinate system.

    Args:
        R_main (3x3 list): Main rotation matrix
        T_main (3x1 list): Main translation vector
        R_alt (3x3 list): Alternative rotation matrix
        T_alt (3x1 list): Alternative translation vector

    Returns:
        R (3x3 list): Rotation matrix to convert alternative to main
        T (3x1 list): Translation vector to convert alternative to main
    """
    R_main = np.array(R_main)
    T_main = np.array(T_main)
    R_alt = np.array(R_alt)
    T_alt = np.array(T_alt)

    R = np.dot(R_main, np.linalg.inv(R_alt))
    T = T_main - np.dot(R, T_alt)

    return R.tolist(), T.tolist()


def chain_transform_matrices(rotation_matrices, translation_vectors):
    """Chain multiple transformations to compute the final transformation from the start to the end of the sequence.

    Args:
        rotation_matrices (list of 3x3 list): List of rotation matrices representing transformations.
        translation_vectors (list of 3x1 list): List of translation vectors representing transformations.

    Returns:
        R_final (3x3 list): Final rotation matrix after all transformations.
        T_final (3x1 list): Final translation vector after all transformations.
    """
    if len(rotation_matrices) != len(translation_vectors):
        raise ValueError("The number of rotation matrices and translation vectors must be the same.")

    R_final = np.eye(3)
    T_final = np.array([[0], [0], [0]])
    for R, T in zip(rotation_matrices, translation_vectors):
        R = np.array(R)
        T = np.array(T)
        print(f'R: {R}')
        print(f'T: {T}')
        R_final = np.dot(R, R_final)
        T_final = np.dot(R, T_final) + T

    print(f'R final: {R_final}')
    print(f'T final: {T_final}')

    return R_final.tolist(), T_final.tolist()


def average_transform_matrices(rotation_matrices, translation_matrices):
    """Average the transform matrices.

    Args:
        rotation_matrices (list of 3x3 list): List of rotation matrices
        translation_matrices (list of 3x1 list): List of translation vectors

    Returns:
        avg_rot_mat (3x3 list): Average rotation matrix
        avg_trans_mat (3x1 list): Average translation vector
    """
    rotation_matrices = np.array(rotation_matrices)
    translation_matrices = np.array(translation_matrices)

    print(f'R shape: {rotation_matrices.shape}')
    print(f'T shape: {translation_matrices.shape}')

    # Average the translation vectors and rotation matrices
    avg_trans_mat = np.mean(translation_matrices, axis=0)
    avg_rot_mat = average_rotation_matrices(rotation_matrices)

    return avg_rot_mat.tolist(), avg_trans_mat.tolist()


def average_rotation_matrices(rot_matrices):
    """Calculate the average rotation matrix from a list of rotation matrices.

    Args:
        rot_matrices (np.ndarray): A list of 3x3 rotation matrices.

    Returns:
        np.ndarray: A 3x3 average rotation matrix calculated from the input list of rotation matrices.
    """
    quaternions = []

    # Convert each rotation matrix to quaternion
    for rot_matrix in rot_matrices:
        rot = Rotation.from_matrix(rot_matrix)
        quat = rot.as_quat()
        quaternions.append(quat)

    # Average quaternions and normalize it
    avg_quat = np.mean(quaternions, axis=0)
    avg_quat = avg_quat / np.linalg.norm(avg_quat)

    # Convert back to a rotation matrix
    avg_rot = Rotation.from_quat(avg_quat)
    avg_rot_matrix = avg_rot.as_matrix()

    return avg_rot_matrix


def find_transformation_paths(data, start, end, path=None):
    """Recursively find all possible transformation paths from start to end.

    Args:
        data (dict): Dictionary containing transformation data
        start (str): Starting coordinate system
        end (str): Ending coordinate system
        path (list): List of keys representing the path taken so far

    Returns:
        paths (list): List of all possible transformation paths from start to end

    Example:
        transform_data = {
            'a-m': {'R': [[1]], 'T': [[1]]},  # Direct transformation from 'a' to 'm'
            'b-c': {'R': [[1]], 'T': [[1]]},  # Transformation from 'b' to 'c'
            'c-b': {'R': [[1]], 'T': [[1]]},  # Transformation from 'c' to 'b'
            'b-a': {'R': [[1]], 'T': [[1]]},  # Transformation from 'b' to 'a'
            'c-m': {'R': [[1]], 'T': [[1]]},  # Direct transformation from 'c' to 'm'
            'e-b': {'R': [[1]], 'T': [[1]]}  # Transformation from 'e' to 'b'
        }
        Paths from b to m:
        ['b-c', 'c-b', 'b-a', 'a-m']
        ['b-c', 'c-m']
        ['b-a', 'a-m']
    """
    if path is None:
        path = []
    if start == end:
        return [path]
    else:
        paths = []
        for key in data:
            prefix, suffix = key.split('-')
            if prefix == start and key not in path:
                new_paths = find_transformation_paths(data, suffix, end, path + [key])
                for p in new_paths:
                    paths.append(p)
        return paths


def compute_paths_transformation(data, paths):
    """Compute and average the transformation matrices from the start to the end of the sequence for each path.

    This function calculates the final rotation and translation matrices for each path, then averages these matrices
    to provide a single representative transformation for each starting system.

    Args:
        data (dict): Dictionary containing transformation data.
        paths (list of str list): List of all possible transformation paths from start to end.

    Returns:
        R_avg (3x3 list): Averaged rotation matrix after combining all paths.
        T_avg (3x1 list): Averaged translation vector after combining all paths.
    """
    R_list = []
    T_list = []

    for path in paths:
        print(f'Path: {path}')
        rotation_matrices = [data[key]['R'] for key in path]
        translation_vectors = [data[key]['T'] for key in path]
        R_final, T_final = chain_transform_matrices(rotation_matrices, translation_vectors)

        R_list.append(R_final)
        T_list.append(T_final)

    R_avg, T_avg = average_transform_matrices(R_list, T_list)

    return R_avg, T_avg


def export_main_transformations_json(input_file, output_file, main_system='m'):
    """Processes transformation data from the input JSON file to compute transformation matrices from all listed
    coordinate systems to a designated main coordinate system, and saves these transformations in a new JSON file.

    Args:
        input_file (str): Path to the input JSON file containing transformation matrices.
        output_file (str): Path where the resultant JSON file with transformations to the main system will be saved.
        main_system (str, optional): The coordinate system to which all transformations will be computed. Defaults to 'm'.

    Outputs:
        A JSON file containing transformation matrices from each available coordinate system to the main system specified.

    Example:
        If the input JSON contains transformations for 'a-b', 'b-c', and 'c-m', calling this function with the main_system
        set to 'm' will produce an output JSON with keys 'a', 'b', 'c' assuming it is possible to calculate these transformations
        based on the data provided.
    """
    with open(input_file, 'r') as file:
        data = json.load(file)

    results = {}
    # Find all starting systems, except the main system
    starts = {key.split('-')[0] for key in data.keys()}
    for start in starts:
        if start != main_system:
            # Find all paths leading to the main system
            paths = find_transformation_paths(data, start, main_system)
            # Compute transformation matrices for these paths
            R, T = compute_paths_transformation(data, paths)
            results[start] = {'R': R, 'T': T}

    # Save the computed transformations to the output file
    with open(output_file, 'w') as file:
        json.dump(results, file, indent=4)

    print(f'Transformations from all systems to the main system {main_system} saved to {output_file}')



import configparser
import inspect
import json
import os

import cv2
import numpy as np
from pupil_apriltags import Detector

from openmmla_vision.utils.influx_client import InfluxDBClientWrapper
from openmmla_vision.utils.input_utils import get_bucket_name, get_function_base
from openmmla_vision.utils.logger_utils import get_logger
from openmmla_vision.utils.mqtt_client import MQTTClientWrapper
from openmmla_vision.utils.relations_utils import is_tag_looking_at_another_2d, get_outward_normal_2d
from openmmla_vision.utils.stream_utils import WebcamVideoStream


class VideoBase:
    """Video base class for AprilTag detection from video stream."""
    logger = get_logger('video-base')

    def __init__(self, project_dir: str = None, config_path: str = None, graphics: bool = True, record: bool = False):
        """Initialize the video base.

        Args:
            project_dir: path to the project directory
            config_path: path to the configuration file
            graphics: whether to show graphics
            record: whether to record video frames
        """
        # Set the project root directory
        if project_dir is None:
            caller_frame = inspect.stack()[1]
            caller_module = inspect.getmodule(caller_frame[0])
            if caller_module is None:
                self.project_dir = os.getcwd()
            else:
                self.project_dir = os.path.dirname(os.path.abspath(caller_module.__file__))
        else:
            self.project_dir = project_dir

        # Determine the configuration path
        if config_path:
            if os.path.isabs(config_path):
                self.config_path = config_path
            else:
                self.config_path = os.path.join(self.project_dir, config_path)
        else:
            self.config_path = os.path.join(self.project_dir, 'conf/video_base.ini')

        # Check if the configuration file exists
        if not os.path.exists(self.config_path):
            raise FileNotFoundError(f"Configuration file not found at {self.config_path}")

        self.graphics = graphics
        self.record = record

        # Set directories
        self.recordings_dir = os.path.join(self.project_dir, 'recordings')
        self.camera_sync_dir = os.path.join(self.project_dir, 'camera_sync')
        self.camera_calib_dir = os.path.join(self.project_dir, 'camera_calib')
        os.makedirs(self.recordings_dir, exist_ok=True)
        os.makedirs(self.camera_sync_dir, exist_ok=True)
        os.makedirs(self.camera_calib_dir, exist_ok=True)

        self.max_badge_id = 12
        self.camera_info = {}
        self.camera_seed = None
        self.sender_id = None
        self.bucket_name = None
        self.video_stream = None
        self.transformation_id = None
        self.transform_matrices_dict = None
        self.camera_configured = False

        self._load_config()
        self._config_objects()

    def _load_config(self):
        self.config = configparser.ConfigParser()
        self.config.read(self.config_path)
        self.tag_size = float(self.config['DEFAULT']['tag_size'])
        self.families = self.config['DEFAULT']['families']
        self.res_width = int(self.config['DEFAULT']['res_width'])
        self.res_height = int(self.config['DEFAULT']['res_height'])
        self.res = (self.res_width, self.res_height)

    def _config_objects(self):
        self.influx_client = InfluxDBClientWrapper(self.config_path)
        self.mqtt_client = MQTTClientWrapper(self.config_path)
        self.detector = Detector(families=self.families, nthreads=4)

    def run(self):
        print('\033]0;Video Base\007')
        func_map = {1: self._start, 2: self._set_camera}

        while True:
            try:
                select_fun = get_function_base()
                if select_fun == 0:
                    self.logger.info("Exiting video base...")
                    break
                func_map.get(select_fun, lambda: print("Invalid option."))()
            except (Exception, KeyboardInterrupt) as e:
                self.logger.warning(
                    f"During running the video base, catch: {'KeyboardInterrupt' if isinstance(e, KeyboardInterrupt) else e}, Come back to the main menu.",
                    exc_info=True)

    def _start(self):
        """Start AprilTag detection"""
        if not self.camera_configured:
            self.logger.warning("Camera is not configured.")
            return self._set_camera()

        # Bucket selection
        self.bucket_name = get_bucket_name(self.influx_client)

        # MQTT client reinitialization
        self.mqtt_client.reinitialise()
        self.mqtt_client.loop_start()

        # Start video stream
        self._configure_video_capture(self.camera_seed)

        try:
            self._process_frames()
        except (Exception, KeyboardInterrupt) as e:
            self.logger.warning("%s, capture interrupted.", e, exc_info=False)
        finally:
            self.video_stream.stop()
            self.mqtt_client.loop_stop()
            if self.graphics:
                cv2.destroyWindow(f'AprilTags Detection from camera {self.sender_id}')
                cv2.waitKey(1)
            # time.sleep(5) # Wait for the threads to terminate

    def _set_camera(self):
        self.camera_configured = False

        self.transform_matrices_dict = self._load_transform_matrices()
        if self.transform_matrices_dict is None:
            self.logger.warning("No transformation matrices found,  please do the camera sync first.")
            return

        self.camera_info = self._configure_camera_params()
        if self.camera_info is None:
            self.logger.warning("Camera configuration failed.")
            return

        available_seeds = self._detect_video_seeds()
        self.camera_seed = self._choose_camera_seed(available_seeds)
        if self.camera_seed is None:
            self.logger.warning("No available camera seed found.")
            return

        self.sender_id = self._choose_sender_id()
        self.camera_configured = True

    def _configure_video_capture(self, camera_seed):
        if self.record:
            save_path = os.path.join(self.recordings_dir, f'{self.bucket_name}/{self.sender_id}')
        else:
            save_path = ''
        self.video_stream = WebcamVideoStream(save_path=save_path, format='MJPG', src=camera_seed, res=(1920, 1080))
        self.video_stream.start()

    def _process_frames(self):
        while True:
            frame = self.video_stream.read()
            if self.camera_info.get("fisheye", False):
                frame = cv2.remap(frame, self.camera_info["map_1"], self.camera_info["map_2"],
                                  interpolation=cv2.INTER_LINEAR, borderMode=cv2.BORDER_CONSTANT)

            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            results = self.detector.detect(gray, estimate_tag_pose=True, camera_params=self.camera_info["params"],
                                           tag_size=self.tag_size)

            tags = {}
            tag_relations = {}

            for tag in results:
                if int(tag.tag_id) > self.max_badge_id:
                    continue

                corners = np.int32(tag.corners)
                normal, tag = get_outward_normal_2d(tag)
                tags[tag.tag_id] = [list(tag.pose_R.tolist()), list(tag.pose_t.tolist())]

                tag_relations.setdefault(tag.tag_id, [])
                for other_tag in results:
                    if other_tag != tag and other_tag.tag_id not in tag_relations[tag.tag_id]:
                        if is_tag_looking_at_another_2d(tag, other_tag, angle_threshold=-0.94,
                                                        distance_threshold=0.5):
                            tag_relations[tag.tag_id].append(str(other_tag.tag_id))

                # Drawing annotations
                if self.graphics:
                    tag_center = np.mean(corners, axis=0)
                    arrow_dir = normal[:2]
                    scale_factor = 50
                    end_point = tag_center + scale_factor * arrow_dir
                    cv2.arrowedLine(frame, tuple(np.int32(tag_center)), tuple(np.int32(end_point)), (0, 0, 255), 2)
                    cv2.polylines(frame, [corners], True, (0, 255, 0), thickness=2)
                    cv2.putText(frame, str(tag.tag_id), org=(int(tag_center[0]) + 10, int(tag_center[1]) + 10),
                                fontFace=cv2.FONT_HERSHEY_SIMPLEX, fontScale=0.8, color=(0, 255, 0), thickness=2)
                    cv2.putText(frame, f"Rot: {tag.pose_R}",
                                (tag.corners[0][0].astype(int), tag.corners[0][1].astype(int) - 30),
                                cv2.FONT_HERSHEY_SIMPLEX, 0.7, (255, 0, 0), 2)
                    cv2.putText(frame, f"Trans: {tag.pose_t}",
                                (tag.corners[0][0].astype(int), tag.corners[0][1].astype(int) - 60),
                                cv2.FONT_HERSHEY_SIMPLEX, 0.7, (255, 0, 0), 2)

            if self.graphics:
                display_frame = cv2.resize(frame, (960, 540))
                cv2.imshow(f'AprilTags Detection from camera {self.sender_id}', display_frame)
                # cv2.imshow(f'AprilTags Detection from camera {self.sender_id}', frame)

            message = {
                "sender_id": self.sender_id,
                "tags": tags,
                "tag_relations": tag_relations,
            }
            message_str = json.dumps(message)
            self.mqtt_client.publish(f'{self.bucket_name}/video', message_str, qos=0, retain=False)

            if cv2.waitKey(1) & 0xFF == ord('q'):
                break

    def _load_transform_matrices(self):
        transformation_choices = [d for d in os.listdir(self.camera_sync_dir) if
                                  d.startswith('transformation_matrices_')]
        for idx, choice in enumerate(transformation_choices):
            print(f"{idx}: {choice}")

        if not transformation_choices:
            return None

        while True:
            try:
                selection = int(input("Choose your main transformation matrices with number: "))
                if not 0 <= selection < len(transformation_choices):
                    self.logger.warning("Invalid selection. Please choose a valid number.")
                else:
                    chosen_transformation = transformation_choices[selection]
                    self.transformation_id = chosen_transformation.split('_')[-1].split('.')[0]
                    break
            except ValueError:
                self.logger.warning("Please enter a valid number.")

        with open(os.path.join(self.camera_sync_dir, chosen_transformation), 'r') as file:
            return json.load(file)

    def _configure_camera_params(self):
        cameras_dir = os.path.join(self.camera_calib_dir, 'cameras')
        camera_choices = [d for d in os.listdir(cameras_dir) if os.path.isdir(os.path.join(cameras_dir, d))]
        for idx, choice in enumerate(camera_choices):
            print(f"{idx}: {choice}")

        if not camera_choices:
            return None

        while True:
            try:
                selection = int(input("Choose your camera name with number: "))
                if not 0 <= selection < len(camera_choices):
                    self.logger.warning("Invalid selection. Please choose a valid number.")
                else:
                    chosen_camera = camera_choices[selection]
                    break
            except ValueError:
                self.logger.warning("Please enter a valid number.")

        fisheye = self.config.getboolean(chosen_camera, 'fisheye')
        params = json.loads(self.config.get(chosen_camera, 'params'))
        camera_info = {"fisheye": fisheye, "params": params, "res": self.res}

        if fisheye:
            K = np.array(json.loads(self.config.get(chosen_camera, 'K')))
            D = np.array(json.loads(self.config.get(chosen_camera, 'D')))
            map_1, map_2 = cv2.fisheye.initUndistortRectifyMap(K, D, np.eye(3), K, self.res, cv2.CV_16SC2)
            camera_info.update({"K": K, "D": D, "map_1": map_1, "map_2": map_2})

        return camera_info

    def _choose_camera_seed(self, available_video_seeds):
        if not available_video_seeds:
            return None
        while True:
            try:
                seed_id = int(input("Choose your video seed id: "))
                if 0 <= seed_id < len(available_video_seeds):
                    return available_video_seeds[seed_id]
                else:
                    self.logger.warning("Invalid selection. Please choose a valid video seed.")
            except ValueError:
                self.logger.warning("Please enter a valid number.")

    def _choose_sender_id(self):
        """Prompt user to choose a sender ID based on available keys in transform_matrices.json configuration."""
        print(f"Available sender ids:\n- {self.transformation_id}")
        for key in self.transform_matrices_dict.keys():
            print(f"- {key}")
        while True:
            sender_id = input("Enter your sender id: ")
            if sender_id in self.transform_matrices_dict or sender_id == self.transformation_id:
                return sender_id
            else:
                print("Invalid selection. Please enter a valid sender id.")

    def _detect_video_seeds(self):
        available_video_seeds = []
        number_of_detected_seeds = 0
        for i in range(4):
            cap = cv2.VideoCapture(i)
            if cap.isOpened():
                print(f"{number_of_detected_seeds} : Camera seed {i} is available.")
                number_of_detected_seeds += 1
                available_video_seeds.append(i)
            cap.release()

        if 'RTMP' in self.config and 'video_streams' in self.config['RTMP']:
            video_stream_list = self.config['RTMP']['video_streams'].split(',')
            for streams in video_stream_list:
                if streams:
                    print(f"{number_of_detected_seeds} : RTMP stream {streams} is available.")
                    available_video_seeds.append(streams)
                    number_of_detected_seeds += 1

        return available_video_seeds



import configparser

import paho.mqtt.client as mqtt


class MQTTClientWrapper(mqtt.Client):
    """Extended MQTT client that loads configuration from a file and adds custom functionalities."""

    def __init__(self, config_path, user_data=None, on_message=None, topics=None):
        """Initialize MQTT client with configurations and optional user data, message callback, and topic
        subscription."""
        super().__init__(callback_api_version=mqtt.CallbackAPIVersion.VERSION2)

        self.config_path = config_path
        self.current_topics = []

        config = configparser.ConfigParser()
        config.read(config_path)
        self.connect(config['MQTT']['mqtt_host'], int(config['MQTT']['mqtt_port']), 60)

        if user_data:
            self.user_data_set(user_data)
        if on_message:
            self.on_message = on_message
        if topics:
            self.subscribe(topics)

    def subscribe(self, topics, qos=0, options=None, properties=None):
        """Override: subscribe to new topics, ensuring previous subscriptions are removed."""
        if isinstance(topics, str):
            topics = [(topics, qos)]
        elif isinstance(topics, tuple):
            topics = [topics]

        if self.current_topics:
            self.unsubscribe(self.current_topics)
        for topic, qos in topics:
            super().subscribe(topic, qos)
        self.current_topics = topics

    def unsubscribe_current(self):
        """Unsubscribe from all currently subscribed topics."""
        if self.current_topics:
            for topic, _ in self.current_topics:
                self.unsubscribe(topic)
            self.current_topics = []

    def reinitialise(self, user_data=None, on_message=None, topics=None):
        """Override: reinitialize the client with new user data, on_message callback, and optionally change the topic
        subscription."""
        self._reset_sockets()
        self.__init__(config_path=self.config_path, user_data=user_data, on_message=on_message, topics=topics)
